// BasicDlg.h : Declaration of the CBasicDlg

#ifndef __BASICDLG_H_
#define __BASICDLG_H_

#include "resource.h"       // main symbols
#include <atlhost.h>

#include "ScriptManager.h"

#ifdef WIN64
	typedef LONG_PTR WW_LONG_PTR;
#else
	typedef LONG WW_LONG_PTR;
#endif

/////////////////////////////////////////////////////////////////////////////
// CBasicDlg
class CBasicDlg : 
	public CAxDialogImpl<CBasicDlg>,
	public IDispEventImpl<IDC_BASICIDECTL1, CBasicDlg>
{
	HTREEITEM current_tree_item;
public:
	CBasicDlg() :
		basicIdeCtlPtr_(0)
	{
	}

	~CBasicDlg()
	{
	}

	enum { IDD = IDD_BASICDLG };

	CAxWindow basicIdeCtl1_;
	_IBasicIdeCtlPtr basicIdeCtlPtr_;
	CAxWindow tree1_;
	CAxWindow button1_;
	CAxWindow check1_;
	std::wstring root_;
	CComPtr<ScriptManager> script_manager_;
	std::map<std::wstring, HTREEITEM> tree_items_;

BEGIN_MSG_MAP(CBasicDlg)
	MESSAGE_HANDLER(WM_INITDIALOG, OnInitDialog)
	MESSAGE_HANDLER(WM_CLOSE, OnClose)
	MESSAGE_HANDLER(WM_DESTROY, OnDestroy)
	MESSAGE_HANDLER(WM_TIMER, OnTimer)
	MESSAGE_HANDLER(WM_SIZE, OnSize)
	NOTIFY_HANDLER(IDC_TREE1, TVN_DELETEITEM, OnTree1_NotifyDeleteItem)
	NOTIFY_HANDLER(IDC_TREE1, TVN_SELCHANGED, OnTree1_NotifySelectChanged)
	COMMAND_HANDLER(IDC_BUTTON1, BN_CLICKED, OnBnClickedButton1)
	COMMAND_HANDLER(IDC_CHECK1, BN_CLICKED, OnBnClickedCheck1)
END_MSG_MAP()
// Handler prototypes:
//  LRESULT MessageHandler(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled);
//  LRESULT CommandHandler(WORD wNotifyCode, WORD wID, HWND hWndCtl, BOOL& bHandled);
//  LRESULT NotifyHandler(int idCtrl, LPNMHDR pnmh, BOOL& bHandled);

	LRESULT OnInitDialog(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		basicIdeCtl1_.Attach(GetDlgItem(IDC_BASICIDECTL1));
		basicIdeCtl1_.QueryControl(&basicIdeCtlPtr_);
		tree1_.Attach(GetDlgItem(IDC_TREE1));
		button1_.Attach(GetDlgItem(IDC_BUTTON1));
		check1_.Attach(GetDlgItem(IDC_CHECK1));

		// advise (connect sinks)
		AtlAdviseSinkMap<CBasicDlg>(this,TRUE);

	#ifdef _DEBUG
		basicIdeCtlPtr_->Trace(wbtAll);
	#endif

		wchar_t szFilename[MAX_PATH];
		const size_t max = sizeof(szFilename)/sizeof(szFilename[0]);
		GetModuleFileName(0,szFilename,max);

		while (true)
		{
			wchar_t* p = wcsrchr(szFilename,L'\\');
			wcscpy_s(p,max-(p-szFilename),L"\\Nodes");
			if (PathFileExists(szFilename))
				break;

			*p = L'\0';
		}

		root_ = szFilename;

		// initialize the script manager
		CComObject<ScriptManager>* script_manager = 0;
		CComObject<ScriptManager>::CreateInstance(&script_manager);
		script_manager->AddRef();
		script_manager_.Attach(script_manager);

		// script manager callback for updating list view as script status changes
		using namespace std::placeholders;
		ScriptManager::script_loaded_status_changed_callback callback =
			std::bind(&CBasicDlg::ScriptLoadedStatusChanged, this, std::placeholders::_1, std::placeholders::_2);
		script_manager_->Attach(basicIdeCtlPtr_, root_.c_str(), callback);

		// initialize the tree view and script names
		TVINSERTSTRUCT tvins;
		tvins.hParent = NULL; // top most level Item
		tvins.hInsertAfter = TVI_ROOT;
		tvins.item.mask = TVIF_TEXT | TVIF_PARAM; 
		tvins.item.iImage = 0;
		tvins.item.iSelectedImage = 0;
		tvins.item.pszText = L"\\";
		tvins.item.lParam = 0;
		HTREEITEM tree_item = TreeView_InsertItem(tree1_, &tvins);

		InitializeTreeView(root_.c_str(), tree_item);

		TreeView_SelectItem(tree1_, tree_item);
		return 1;  // Let the system set the focus
	}

	void InitializeTreeView(const wchar_t* dir, HTREEITEM tree_item)
	{
		WIN32_FIND_DATA ffd;
		std::wstring file(dir);
		file += L"\\*.*";
		HANDLE hFind = FindFirstFile(file.c_str(), &ffd);
		if (hFind != INVALID_HANDLE_VALUE)
		{
			do
			{
				if (!(ffd.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY))
				{
					std::wstring* path = new std::wstring(dir+root_.size());
					*path += L"\\";
					*path += ffd.cFileName;
					std::wstring name = ffd.cFileName;
					name += L" ";
					TVINSERTSTRUCT tvins;
					tvins.hParent = tree_item;
					tvins.hInsertAfter = TVI_LAST;
					tvins.item.mask = TVIF_TEXT | TVIF_PARAM; 
					tvins.item.iImage = 0;
					tvins.item.iSelectedImage = 0;
					tvins.item.pszText = (wchar_t*)name.c_str();
					tvins.item.lParam = (LPARAM)path;
					HTREEITEM child_tree_item = TreeView_InsertItem(tree1_, &tvins);
					tree_items_.emplace(*path, child_tree_item);
				}
			} while (FindNextFile(hFind, &ffd));
		}

		hFind = FindFirstFile(file.c_str(), &ffd);
		if (hFind != INVALID_HANDLE_VALUE)
		{
			do
			{
				if ((ffd.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY) && ffd.cFileName[0] != L'.')
				{
					std::wstring name = ffd.cFileName;
					name += L"\\";
					TVINSERTSTRUCT tvins;
					tvins.hParent = tree_item;
					tvins.hInsertAfter = TVI_LAST;
					tvins.item.mask = TVIF_TEXT | TVIF_PARAM; 
					tvins.item.iImage = 0;
					tvins.item.iSelectedImage = 0;
					tvins.item.pszText = (wchar_t*)name.c_str();
					tvins.item.lParam = 0;
					HTREEITEM child_tree_item = TreeView_InsertItem(tree1_, &tvins);
					std::wstring child_dir = dir;
					child_dir += L"\\";
					child_dir += ffd.cFileName;
					InitializeTreeView(child_dir.c_str(), child_tree_item);
				}
			} while (FindNextFile(hFind, &ffd));
		}

		TreeView_Expand(tree1_, tree_item, TVE_EXPAND);
	}

	void ScriptLoadedStatusChanged(const std::wstring& script_name, bool script_loaded_status)
	{
		auto it = tree_items_.find(script_name);
		if (it != tree_items_.end())
		{
			WCHAR szText[100];
			TVITEM tvitem;
			tvitem.mask = TVIF_HANDLE | TVIF_TEXT;
			tvitem.hItem = it->second;
			tvitem.pszText = szText;
			tvitem.cchTextMax = sizeof(szText)/sizeof(WCHAR);
			TreeView_GetItem(tree1_, &tvitem);
			size_t len = wcslen(szText);
			szText[len-1] = script_loaded_status ? L'+' : L' ';
			TreeView_SetItem(tree1_, &tvitem);
		}
	}

	LRESULT OnClose(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		short iResult = 0;
		basicIdeCtlPtr_->Shutdown(&iResult);
		switch (iResult)
		{
		case -1: // WinWrap Basic execution is nested.
			// this case is not possible when Synchronized is True
			// SetTimer(1, 500);
			return 0; // can't close now (try again later)
		case 0: // Shutdown okay.
			break;
		case 1: // Shutdown canceled by the user.
			return 0;
		}

		// make sure that all macros are halted
		if (!script_manager_->Shutdown())
		{
			SetTimer(1, 500);
			return 0; // can't close now (try again later)
		}

		// free script manager resources
		script_manager_->Disconnect();
		script_manager_.Release();

		EndDialog(IDOK);
		return 0;
	}

	LRESULT OnDestroy(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		if (basicIdeCtlPtr_)
		{
			// unadvise (disconnect sinks)
			AtlAdviseSinkMap<CBasicDlg>(this,FALSE);

			VARIANT_BOOL bResult;
			basicIdeCtlPtr_->Disconnect(&bResult);
			basicIdeCtlPtr_.Release();
			basicIdeCtl1_.Detach();
		}

		return 0;
	}

	LRESULT OnSize(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		if (!IsIconic())
		{
			CWindow wnd = GetDesktopWindow();
			RECT rc;
			basicIdeCtl1_.GetWindowRect(&rc);
			wnd.MapWindowPoints(*this,&rc);
			basicIdeCtl1_.MoveWindow(rc.left,0,LOWORD(lParam)-rc.left,HIWORD(lParam));
			tree1_.GetWindowRect(&rc);
			wnd.MapWindowPoints(*this,&rc);
			tree1_.MoveWindow(rc.left,rc.top,rc.right-rc.left,HIWORD(lParam)-rc.top-20);
		}

		return 0;
	}

	LRESULT OnTimer(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		switch (wParam)
		{
		case 1:
			// *** AddBasic: deferred close
			{
				VARIANT_BOOL b = VARIANT_FALSE;
				basicIdeCtlPtr_->get_InEvent(&b);
				if (!b)
				{
					PostMessage(WM_CLOSE, 0, 0l);
					KillTimer(1);
				}
			}

			break;
			// ***
		}

		return 0;
	}

	LRESULT OnTree1_NotifyDeleteItem(WPARAM wParam, LPNMHDR lParam, BOOL& bHandled)
	{
		auto pnmtv = (LPNMTREEVIEW)lParam;
		auto tvitem = pnmtv->itemOld;
		std::wstring* pstr = (std::wstring*)tvitem.lParam;
		delete pstr;
		return 0;
	}

	LRESULT OnTree1_NotifySelectChanged(WPARAM wParam, LPNMHDR lParam, BOOL& bHandled)
	{
		auto pnmtv = (LPNMTREEVIEW)lParam;
		auto tvitem = pnmtv->itemNew;
		current_tree_item = tvitem.hItem;
		auto name = (const std::wstring*)tvitem.lParam;
		if (name != 0)
		{
			// view code into the IDE
			basicIdeCtlPtr_->put_FileName(0, bstr_t(name->c_str()));
		}
		else
		{
            // no code to view
            basicIdeCtlPtr_->put_FileName(0, bstr_t(L"<none>"));
		}

		return 0;
	}
	LRESULT OnBnClickedCheck1(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
	{
		bool fast = check1_.SendMessage(BM_GETCHECK) != 0;
		script_manager_->Fast(fast);
		return 0;
	}
	LRESULT OnBnClickedButton1(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
	{
		// clear the immediate window
		{
			CComPtr<Result> result;
			LPSAFEARRAY psa = 0;
			basicIdeCtlPtr_->Query(bstr_t(L"Set IDE Immediate"), &psa, &result);
		}

		// run all the script in the sub tree

		auto script_names = ScriptNames(current_tree_item);
		variant_t default_value;
		default_value = L"<null>";
		std::list<variant_t> results;
		script_manager_->EvaluateScripts(script_names, results, &default_value);

		// show the results in the immediate window
		std::wstring buffer;
		for (const auto& result : results)
		{
			variant_t v(result);
			v.ChangeType(VT_BSTR);
			buffer += v.bstrVal;
			buffer += L"\r\n";
		}

		CComPtr<Result> result;
		LPSAFEARRAY psa = SafeArrayCreateVector(VT_VARIANT,0,1);
		long index = 0;
		VARIANT* v = 0;
		SafeArrayAccessData(psa,(void**)&v);
		v->vt = VT_BSTR;
		v->bstrVal = SysAllocString(buffer.c_str());
		SafeArrayUnaccessData(psa);
		basicIdeCtlPtr_->Query(bstr_t(L"Set IDE Immediate"), &psa, &result);
		SafeArrayDestroy(psa);
		return 0;
	}
	std::list<std::wstring> ScriptNames(HTREEITEM tree_item)
    {
        // return the list of script nodes at and under this tree node
        auto script_names = std::list<std::wstring>();
        BuildScriptNameList(tree_item, script_names);
        return script_names;
    }

    void BuildScriptNameList(HTREEITEM tree_item, std::list<std::wstring>& script_names)
    {
        // recursive for one tree node
		TVITEM tvitem;
		tvitem.mask = TVIF_HANDLE;
		tvitem.hItem = tree_item;
		TreeView_GetItem(tree1_, &tvitem);
		auto script_name = (const std::wstring*)tvitem.lParam;
        if (script_name == 0)
            BuildChildrenScriptNameList(tree_item, script_names);
        else
            script_names.push_back(*script_name);
    }

    void BuildChildrenScriptNameList(HTREEITEM hParentItem, std::list<std::wstring>& script_names)
    {
        // recursive for all child tree nodes
		HTREEITEM hChildItem = TreeView_GetChild(tree1_,hParentItem);
		while (hChildItem)
		{
			BuildScriptNameList(hChildItem, script_names);
			hChildItem = TreeView_GetNextSibling(tree1_,hChildItem);
		}
    }
public :

BEGIN_SINK_MAP(CBasicDlg)
	//Make sure the Event Handlers have __stdcall calling convention
	SINK_ENTRY(IDC_BASICIDECTL1, 0x1b, OnShowFormBasicidectl1)
	SINK_ENTRY(IDC_BASICIDECTL1, 0x1c, OnDebugTraceBasicidectl1)
	SINK_ENTRY(IDC_BASICIDECTL1, 0x1e, OnOverrideModalWindowOwnerBasicidectl1)
	SINK_ENTRY(IDC_BASICIDECTL1, 0x1f, OnModalDialogBasicidectl1)
	SINK_ENTRY(IDC_BASICIDECTL1, 0x27, OnModalIdeDialogBasicidectl1)
	SINK_ENTRY(IDC_BASICIDECTL1, 0x3, OnDoEventsBasicidectl1)
	SINK_ENTRY(IDC_BASICIDECTL1, 0x1, OnChangeBasicidectl1)
	SINK_ENTRY(IDC_BASICIDECTL1, 32, EnterDesignModeBasicidectl1)
	SINK_ENTRY(IDC_BASICIDECTL1, 33, LeaveDesignModeBasicidectl1)
	SINK_ENTRY(IDC_BASICIDECTL1, 23, OverrideMenuCommandBasicidectl1)
	SINK_ENTRY(IDC_BASICIDECTL1, 46, SynchronizingBasicidectl1)
END_SINK_MAP()

	VOID __stdcall OnChangeBasicidectl1()
	{
		CComBSTR sCaption;
		GetWindowText(&sCaption);
		CComBSTR sBasicCaption;
		basicIdeCtlPtr_->get_Caption(&sBasicCaption);
		if (sBasicCaption != sCaption)
		{
			USES_CONVERSION;
			SetWindowText(OLE2CT(sBasicCaption));
		}
	}
	VOID __stdcall OnShowFormBasicidectl1()
	{
		// make the top level parent window visible
		if (IsIconic())
			SendMessage(WM_SYSCOMMAND,SC_RESTORE);

		ShowWindow(SW_SHOW);
		SetForegroundWindow(m_hWnd);
		basicIdeCtl1_.SetFocus();
	}
	VOID __stdcall OnDebugTraceBasicidectl1(BSTR * Text)
	{
		USES_CONVERSION;
		ATLTRACE(_T("%s\n"),OLE2CT(*Text));
	}
	VOID __stdcall OnOverrideModalWindowOwnerBasicidectl1(WW_LONG_PTR * hWndOwner)
	{
		// this will be the parent window for a modal dialog instruction
		*hWndOwner = (WW_LONG_PTR)GetLastActivePopup().m_hWnd;
	}
	VOID __stdcall OnModalDialogBasicidectl1(VARIANT_BOOL Displayed)
	{
		// display or hide a modal dialog instruction
		// (InputBox, MsgBox or UserDialog)
//		AfxGetApp()->EnableModeless(!Displayed);
		basicIdeCtlPtr_->put_Enabled(Displayed ? VARIANT_FALSE : VARIANT_TRUE);
	}
	VOID __stdcall OnModalIdeDialogBasicidectl1(VARIANT_BOOL Displayed)
	{
		// display or hide a modal dialog (IDE modal dialog)
		EnableWindow(!Displayed);
	}
	VOID __stdcall OnDoEventsBasicidectl1()
	{
		MSG msg;
		while (PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
		{
			if (msg.message == WM_QUIT)
			{
				PostQuitMessage((int)msg.wParam);
				break;
			}

			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	void __stdcall EnterDesignModeBasicidectl1(BOOL* Cancel)
	{
		button1_.EnableWindow(FALSE);
		check1_.EnableWindow(TRUE);
	}
	void __stdcall LeaveDesignModeBasicidectl1(BOOL* Cancel)
	{
		button1_.EnableWindow(TRUE);
		check1_.EnableWindow(FALSE);
	}
	void __stdcall OverrideMenuCommandBasicidectl1(short CmdId, VARIANT_BOOL* Cancel)
	{
		switch (CmdId)
		{
		case wbcFileOpen:
		case wbcFileClose:
		case wbcFileSaveAs:
		case wbcMacroRun:
			*Cancel = VARIANT_TRUE;
			break;
		}
	}
	void __stdcall SynchronizingBasicidectl1(BSTR* Param, long Id)
	{
		// synchronize the NoUI with the IDE
		script_manager_->SynchronizeNoUI(Param, Id);
	}
};

#endif //__BASICDLG_H_
